"""Tests for certbot.helpful_parser"""
import sys

import pytest

from certbot._internal.cli import HelpfulArgumentParser


class TestScanningFlags:
    '''Test the prescan_for_flag method of HelpfulArgumentParser'''
    def test_prescan_no_help_flag(self):
        arg_parser = HelpfulArgumentParser(['run'], {})
        detected_flag = arg_parser.prescan_for_flag('--help',
                                                        ['all', 'certonly'])
        assert detected_flag is False
        detected_flag = arg_parser.prescan_for_flag('-h',
                                                        ['all, certonly'])
        assert detected_flag is False

    def test_prescan_unvalid_topic(self):
        arg_parser = HelpfulArgumentParser(['--help', 'all'], {})
        detected_flag = arg_parser.prescan_for_flag('--help',
                                                    ['potato'])
        assert detected_flag is True
        detected_flag = arg_parser.prescan_for_flag('-h',
                                                    arg_parser.help_topics)
        assert detected_flag is False

    def test_prescan_valid_topic(self):
        arg_parser = HelpfulArgumentParser(['-h', 'all'], {})
        detected_flag = arg_parser.prescan_for_flag('-h',
                                                    arg_parser.help_topics)
        assert detected_flag == 'all'
        detected_flag = arg_parser.prescan_for_flag('--help',
                                                    arg_parser.help_topics)
        assert detected_flag is False

class TestDetermineVerbs:
    '''Tests for determine_verb methods of HelpfulArgumentParser'''
    def test_determine_verb_wrong_verb(self):
        arg_parser = HelpfulArgumentParser(['potato'], {})
        assert arg_parser.verb == "run"
        assert arg_parser.args == ["potato"]

    def test_determine_verb_help(self):
        arg_parser = HelpfulArgumentParser(['--help', 'everything'], {})
        assert arg_parser.verb == "help"
        assert arg_parser.args == ["--help", "everything"]
        arg_parser = HelpfulArgumentParser(['-d', 'some_domain', '--help',
                                               'all'], {})
        assert arg_parser.verb == "help"
        assert arg_parser.args == ['-d', 'some_domain', '--help',
                                               'all']

    def test_determine_verb(self):
        arg_parser = HelpfulArgumentParser(['certonly'], {})
        assert arg_parser.verb == 'certonly'
        assert arg_parser.args == []

        arg_parser = HelpfulArgumentParser(['auth'], {})
        assert arg_parser.verb == 'certonly'
        assert arg_parser.args == []

        arg_parser = HelpfulArgumentParser(['everything'], {})
        assert arg_parser.verb == 'run'
        assert arg_parser.args == []


class TestAdd:
    '''Tests for add method in HelpfulArgumentParser'''
    def test_add_trivial_argument(self):
        arg_parser = HelpfulArgumentParser(['run'], {})
        arg_parser.add(None, "--hello-world")
        parsed_args = arg_parser.parser.parse_args(['--hello-world',
                                                    'Hello World!'])
        assert parsed_args.hello_world == 'Hello World!'
        assert not hasattr(parsed_args, 'potato')

    def test_add_expected_argument(self):
        arg_parser = HelpfulArgumentParser(['--help', 'run'], {})
        arg_parser.add(
                [None, "run", "certonly", "register"],
                "--eab-kid", dest="eab_kid", action="store",
                metavar="EAB_KID",
                help="Key Identifier for External Account Binding")
        parsed_args = arg_parser.parser.parse_args(["--eab-kid", None])
        assert parsed_args.eab_kid is None
        assert hasattr(parsed_args, 'eab_kid')


class TestAddGroup:
    '''Test add_group method of HelpfulArgumentParser'''
    def test_add_group_no_input(self):
        arg_parser = HelpfulArgumentParser(['run'], {})
        with pytest.raises(TypeError):
            arg_parser.add_group()

    def test_add_group_topic_not_visible(self):
        # The user request help on run. A topic that given somewhere in the
        # args won't be added to the groups in the parser.
        arg_parser = HelpfulArgumentParser(['--help', 'run'], {})
        arg_parser.add_group("auth",
                                         description="description of auth")
        assert arg_parser.groups == {}

    def test_add_group_topic_requested_help(self):
        arg_parser = HelpfulArgumentParser(['--help', 'run'], {})
        arg_parser.add_group("run",
                                         description="description of run")
        assert arg_parser.groups["run"]
        arg_parser.add_group("certonly", description="description of certonly")
        with pytest.raises(KeyError):
            assert arg_parser.groups["certonly"] is False


if __name__ == '__main__':
    sys.exit(pytest.main(sys.argv[1:] + [__file__]))  # pragma: no cover



"""A dummy module with no effect for use on systems without readline."""
from typing import Callable
from typing import Iterable
from typing import Optional


def get_completer() -> Optional[Callable[[], str]]:
    """An empty implementation of readline.get_completer."""


def get_completer_delims() -> list[str]:
    """An empty implementation of readline.get_completer_delims."""
    return []


def parse_and_bind(unused_command: str) -> None:
    """An empty implementation of readline.parse_and_bind."""


def set_completer(unused_function: Optional[Callable[[], str]] = None) -> None:
    """An empty implementation of readline.set_completer."""


def set_completer_delims(unused_delims: Iterable[str]) -> None:
    """An empty implementation of readline.set_completer_delims."""




"""Tests for certbot._internal.cert_manager."""
# pylint: disable=protected-access
import re
import shutil
import sys
import tempfile
import unittest
from unittest import mock

import configobj
import pytest

from certbot import configuration
from certbot import errors
from certbot._internal import san
from certbot._internal.storage import ALL_FOUR
from certbot._internal.tests import storage_test
from certbot.compat import filesystem
from certbot.compat import os
from certbot.display import util as display_util
from certbot.tests import util as test_util


class BaseCertManagerTest(test_util.ConfigTestCase):
    """Base class for setting up Cert Manager tests.
    """
    def setUp(self):
        super().setUp()

        self.config.quiet = False
        filesystem.makedirs(self.config.renewal_configs_dir)

        self.domains = {
            "example.org": None,
            "other.com": os.path.join(self.config.config_dir, "specialarchive")
        }
        self.config_files = {domain: self._set_up_config(domain, self.domains[domain])
            for domain in self.domains}

        # We also create a file that isn't a renewal config in the same
        # location to test that logic that reads in all-and-only renewal
        # configs will ignore it and NOT attempt to parse it.
        with open(os.path.join(self.config.renewal_configs_dir, "IGNORE.THIS"), "w") as junk:
            junk.write("This file should be ignored!")

    def _set_up_config(self, domain, custom_archive):
        # TODO: maybe provide NamespaceConfig.make_dirs?
        # TODO: main() should create those dirs, c.f. #902
        filesystem.makedirs(os.path.join(self.config.live_dir, domain))
        config_file = configobj.ConfigObj()

        if custom_archive is not None:
            filesystem.makedirs(custom_archive)
            config_file["archive_dir"] = custom_archive
        else:
            filesystem.makedirs(os.path.join(self.config.default_archive_dir, domain))

        for kind in ALL_FOUR:
            config_file[kind] = os.path.join(self.config.live_dir, domain,
                                        kind + ".pem")

        config_file.filename = os.path.join(self.config.renewal_configs_dir,
                                       domain + ".conf")
        config_file.write()
        return config_file


class DeleteTest(storage_test.BaseRenewableCertTest):
    """Tests for certbot._internal.cert_manager.delete
    """

    def _call(self):
        from certbot._internal import cert_manager
        cert_manager.delete(self.config)

    @test_util.patch_display_util()
    @mock.patch('certbot.display.util.notify')
    @mock.patch('certbot._internal.cert_manager.lineage_for_certname')
    @mock.patch('certbot._internal.storage.delete_files')
    def test_delete_from_config_yes(self, mock_delete_files, mock_lineage_for_certname,
        mock_notify, mock_util):
        """Test delete"""
        mock_lineage_for_certname.return_value = self.test_rc
        mock_util().yesno.return_value = True
        self.config.certname = "example.org"
        self._call()
        mock_delete_files.assert_called_once_with(self.config, "example.org")
        mock_notify.assert_called_once_with(
            "Deleted all files relating to certificate example.org."
        )

    @test_util.patch_display_util()
    @mock.patch('certbot._internal.cert_manager.lineage_for_certname')
    @mock.patch('certbot._internal.storage.delete_files')
    def test_delete_from_config_no(self, mock_delete_files, mock_lineage_for_certname,
        mock_util):
        """Test delete"""
        mock_lineage_for_certname.return_value = self.test_rc
        mock_util().yesno.return_value = False
        self.config.certname = "example.org"
        self._call()
        assert mock_delete_files.call_count == 0

    @test_util.patch_display_util()
    @mock.patch('certbot._internal.cert_manager.lineage_for_certname')
    @mock.patch('certbot._internal.storage.delete_files')
    def test_delete_interactive_single_yes(self, mock_delete_files, mock_lineage_for_certname,
        mock_util):
        """Test delete"""
        mock_lineage_for_certname.return_value = self.test_rc
        mock_util().checklist.return_value = (display_util.OK, ["example.org"])
        mock_util().yesno.return_value = True
        self._call()
        mock_delete_files.assert_called_once_with(self.config, "example.org")

    @test_util.patch_display_util()
    @mock.patch('certbot._internal.cert_manager.lineage_for_certname')
    @mock.patch('certbot._internal.storage.delete_files')
    def test_delete_interactive_single_no(self, mock_delete_files, mock_lineage_for_certname,
        mock_util):
        """Test delete"""
        mock_lineage_for_certname.return_value = self.test_rc
        mock_util().checklist.return_value = (display_util.OK, ["example.org"])
        mock_util().yesno.return_value = False
        self._call()
        assert mock_delete_files.call_count == 0

    @test_util.patch_display_util()
    @mock.patch('certbot._internal.cert_manager.lineage_for_certname')
    @mock.patch('certbot._internal.storage.delete_files')
    def test_delete_interactive_multiple_yes(self, mock_delete_files, mock_lineage_for_certname,
        mock_util):
        """Test delete"""
        mock_lineage_for_certname.return_value = self.test_rc
        mock_util().checklist.return_value = (display_util.OK, ["example.org", "other.org"])
        mock_util().yesno.return_value = True
        self._call()
        mock_delete_files.assert_any_call(self.config, "example.org")
        mock_delete_files.assert_any_call(self.config, "other.org")
        assert mock_delete_files.call_count == 2

    @test_util.patch_display_util()
    @mock.patch('certbot._internal.cert_manager.lineage_for_certname')
    @mock.patch('certbot._internal.storage.delete_files')
    def test_delete_interactive_multiple_no(self, mock_delete_files, mock_lineage_for_certname,
        mock_util):
        """Test delete"""
        mock_lineage_for_certname.return_value = self.test_rc
        mock_util().checklist.return_value = (display_util.OK, ["example.org", "other.org"])
        mock_util().yesno.return_value = False
        self._call()
        assert mock_delete_files.call_count == 0


class CertificatesTest(BaseCertManagerTest):
    """Tests for certbot._internal.cert_manager.certificates
    """
    def _certificates(self, *args, **kwargs):
        from certbot._internal.cert_manager import certificates
        return certificates(*args, **kwargs)

    @mock.patch('certbot._internal.cert_manager.logger')
    @test_util.patch_display_util()
    def test_certificates_parse_fail(self, mock_utility, mock_logger):
        self._certificates(self.config)
        assert mock_logger.warning.called #pylint: disable=no-member
        assert mock_utility.called

    @mock.patch('certbot._internal.cert_manager.logger')
    @test_util.patch_display_util()
    def test_certificates_quiet(self, mock_utility, mock_logger):
        self.config.quiet = True
        self._certificates(self.config)
        assert mock_utility.notification.called is False
        assert mock_logger.warning.called #pylint: disable=no-member

    @mock.patch('certbot.crypto_util.verify_renewable_cert')
    @mock.patch('certbot._internal.cert_manager.logger')
    @test_util.patch_display_util()
    @mock.patch("certbot._internal.storage.RenewableCert")
    @mock.patch('certbot._internal.cert_manager._report_human_readable')
    def test_certificates_parse_success(self, mock_report, mock_renewable_cert,
        mock_utility, mock_logger, mock_verifier):
        mock_verifier.return_value = None
        mock_report.return_value = ""
        self._certificates(self.config)
        assert mock_logger.warning.called is False
        assert mock_report.called
        assert mock_utility.called
        assert mock_renewable_cert.called

    @mock.patch('certbot._internal.cert_manager.logger')
    @test_util.patch_display_util()
    def test_certificates_no_files(self, mock_utility, mock_logger):
        empty_tempdir = tempfile.mkdtemp()
        empty_config = configuration.NamespaceConfig(mock.MagicMock(
            config_dir=os.path.join(empty_tempdir, "config"),
            work_dir=os.path.join(empty_tempdir, "work"),
            logs_dir=os.path.join(empty_tempdir, "logs"),
            quiet=False
        ))

        filesystem.makedirs(empty_config.renewal_configs_dir)
        self._certificates(empty_config)
        assert mock_logger.warning.called is False
        assert mock_utility.called
        shutil.rmtree(empty_tempdir)

    @mock.patch('certbot.crypto_util.get_serial_from_cert')
    @mock.patch('certbot._internal.cert_manager.ocsp.RevocationChecker.ocsp_revoked')
    def test_report_human_readable(self, mock_revoked, mock_serial):
        mock_revoked.return_value = None
        mock_serial.return_value = 1234567890
        import datetime

        from certbot._internal import cert_manager
        expiry = datetime.datetime.now(datetime.timezone.utc)

        cert = mock.MagicMock(lineagename="nameone")
        cert.target_expiry = expiry
        cert.sans.return_value = [san.DNSName("nameone"), san.DNSName("nametwo")]
        cert.is_test_cert = False
        parsed_certs = [cert]

        mock_config = mock.MagicMock(certname=None, lineagename=None)
        mock_config.domains = []
        mock_config.ip_addresses = []
        # pylint: disable=protected-access

        # pylint: disable=protected-access
        def get_report():
            return cert_manager._report_human_readable(mock_config, parsed_certs)

        out = get_report()
        assert "INVALID: EXPIRED" in out

        cert.target_expiry += datetime.timedelta(hours=2)
        # pylint: disable=protected-access
        out = get_report()
        assert ('1 hour' in out or '2 hour(s)' in out) is True
        assert 'VALID' in out
        assert 'INVALID' not in out

        cert.target_expiry += datetime.timedelta(days=1)
        # pylint: disable=protected-access
        out = get_report()
        assert '1 day' in out
        assert 'under' not in out
        assert 'VALID' in out
        assert 'INVALID' not in out

        cert.target_expiry += datetime.timedelta(days=2)
        # pylint: disable=protected-access
        out = get_report()
        assert '3 days' in out
        assert 'VALID' in out
        assert 'INVALID' not in out

        cert.is_test_cert = True
        mock_revoked.return_value = True
        out = get_report()
        assert 'INVALID: TEST_CERT, REVOKED' in out

        cert = mock.MagicMock(lineagename="indescribable")
        cert.target_expiry = expiry
        cert.sans.return_value = [san.DNSName("nameone"), san.DNSName("thrice.named")]
        cert.is_test_cert = True
        parsed_certs.append(cert)

        out = get_report()
        assert len(re.findall("INVALID:", out)) == 2
        mock_config.domains = [san.DNSName("thrice.named")]
        out = get_report()
        assert len(re.findall("INVALID:", out)) == 1
        mock_config.domains = [san.DNSName("nameone")]
        out = get_report()
        assert len(re.findall("INVALID:", out)) == 2
        mock_config.certname = "indescribable"
        out = get_report()
        assert len(re.findall("INVALID:", out)) == 1
        mock_config.certname = "horror"
        out = get_report()
        assert len(re.findall("INVALID:", out)) == 0


class SearchLineagesTest(BaseCertManagerTest):
    """Tests for certbot._internal.cert_manager._search_lineages."""

    @mock.patch('certbot.util.make_or_verify_dir')
    @mock.patch('certbot._internal.storage.renewal_conf_files')
    @mock.patch('certbot._internal.storage.RenewableCert')
    def test_cert_storage_error(self, mock_renewable_cert, mock_renewal_conf_files,
                                mock_make_or_verify_dir):
        mock_renewal_conf_files.return_value = ["badfile"]
        mock_renewable_cert.side_effect = errors.CertStorageError
        from certbot._internal import cert_manager

        # pylint: disable=protected-access
        assert cert_manager._search_lineages(self.config, lambda x: x, "check") == "check"
        assert mock_make_or_verify_dir.called


class LineageForCertnameTest(BaseCertManagerTest):
    """Tests for certbot._internal.cert_manager.lineage_for_certname"""

    @mock.patch('certbot.util.make_or_verify_dir')
    @mock.patch('certbot._internal.storage.renewal_file_for_certname')
    @mock.patch('certbot._internal.storage.RenewableCert')
    def test_found_match(self, mock_renewable_cert, mock_renewal_conf_file,
                         mock_make_or_verify_dir):
        mock_renewal_conf_file.return_value = "somefile.conf"
        mock_match = mock.Mock(lineagename="example.com")
        mock_renewable_cert.return_value = mock_match
        from certbot._internal import cert_manager
        assert cert_manager.lineage_for_certname(self.config, "example.com") == mock_match
        assert mock_make_or_verify_dir.called

    @mock.patch('certbot.util.make_or_verify_dir')
    @mock.patch('certbot._internal.storage.renewal_file_for_certname')
    def test_no_match(self, mock_renewal_conf_file, mock_make_or_verify_dir):
        mock_renewal_conf_file.return_value = "other.com.conf"
        from certbot._internal import cert_manager
        assert cert_manager.lineage_for_certname(self.config, "example.com") is None
        assert mock_make_or_verify_dir.called

    @mock.patch('certbot.util.make_or_verify_dir')
    @mock.patch('certbot._internal.storage.renewal_file_for_certname')
    def test_no_renewal_file(self, mock_renewal_conf_file, mock_make_or_verify_dir):
        mock_renewal_conf_file.side_effect = errors.CertStorageError()
        from certbot._internal import cert_manager
        assert cert_manager.lineage_for_certname(self.config, "example.com") is None
        assert mock_make_or_verify_dir.called


class DomainsForCertnameTest(BaseCertManagerTest):
    """Tests for certbot._internal.cert_manager.sans_for_certname"""

    @mock.patch('certbot.util.make_or_verify_dir')
    @mock.patch('certbot._internal.storage.renewal_file_for_certname')
    @mock.patch('certbot._internal.storage.RenewableCert')
    def test_found_match(self, mock_renewable_cert, mock_renewal_conf_file,
                         mock_make_or_verify_dir):
        mock_renewal_conf_file.return_value = "somefile.conf"
        mock_match = mock.Mock(lineagename="example.com")
        domains = [san.DNSName("example.com"), san.DNSName("example.org")]
        mock_match.sans.return_value = domains
        mock_renewable_cert.return_value = mock_match
        from certbot._internal import cert_manager
        assert cert_manager.sans_for_certname(self.config, "example.com") == \
               domains
        assert mock_make_or_verify_dir.called

    @mock.patch('certbot.util.make_or_verify_dir')
    @mock.patch('certbot._internal.storage.renewal_file_for_certname')
    def test_no_match(self, mock_renewal_conf_file, mock_make_or_verify_dir):
        mock_renewal_conf_file.return_value = "somefile.conf"
        from certbot._internal import cert_manager
        assert cert_manager.sans_for_certname(self.config, "other.com") is None
        assert mock_make_or_verify_dir.called


class DuplicativeCertsTest(storage_test.BaseRenewableCertTest):
    """Test to avoid duplicate lineages."""

    def setUp(self):
        super().setUp()
        self.config_file.write()
        self._write_out_ex_kinds()

    @mock.patch('certbot.util.make_or_verify_dir')
    def test_find_duplicative_names(self, unused_makedir):
        from certbot._internal.cert_manager import find_duplicative_certs
        test_cert = test_util.load_vector('cert-san_512.pem')
        with open(self.test_rc.cert, 'wb') as f:
            f.write(test_cert)

        # No overlap at all
        result = find_duplicative_certs(
            self.config, [san.DNSName('wow.net'), san.DNSName('hooray.org')])
        assert result == (None, None)

        # Totally identical
        result = find_duplicative_certs(
            self.config, [san.DNSName('example.com'), san.DNSName('www.example.com')])
        assert result[0].configfile.filename.endswith('example.org.conf')
        assert result[1] is None

        # Superset
        result = find_duplicative_certs(
            self.config, [san.DNSName('example.com'), san.DNSName('www.example.com'),
                          san.DNSName('something.new')])
        assert result[0] is None
        assert result[1].configfile.filename.endswith('example.org.conf')

        # Partial overlap doesn't count
        result = find_duplicative_certs(
            self.config, [san.DNSName('example.com'), san.DNSName('something.new')])
        assert result == (None, None)


class CertPathToLineageTest(storage_test.BaseRenewableCertTest):
    """Tests for certbot._internal.cert_manager.cert_path_to_lineage"""

    def setUp(self):
        super().setUp()
        self.config_file.write()
        self._write_out_ex_kinds()
        self.fullchain = os.path.join(self.config.config_dir, 'live', 'example.org',
                'fullchain.pem')
        self.config.cert_path = self.fullchain

    def _call(self, cli_config):
        from certbot._internal.cert_manager import cert_path_to_lineage
        return cert_path_to_lineage(cli_config)

    def _archive_files(self, cli_config, filetype):
        from certbot._internal.cert_manager import _archive_files
        return _archive_files(cli_config, filetype)

    def test_basic_match(self):
        assert 'example.org' == self._call(self.config)

    def test_no_match_exists(self):
        bad_test_config = self.config
        bad_test_config.cert_path = os.path.join(self.config.config_dir, 'live',
                'SailorMoon', 'fullchain.pem')
        with pytest.raises(errors.Error):
            self._call(bad_test_config)

    @mock.patch('certbot._internal.cert_manager._acceptable_matches')
    def test_options_fullchain(self, mock_acceptable_matches):
        mock_acceptable_matches.return_value = [lambda x: x.fullchain_path]
        self.config.fullchain_path = self.fullchain
        assert 'example.org' == self._call(self.config)

    @mock.patch('certbot._internal.cert_manager._acceptable_matches')
    def test_options_cert_path(self, mock_acceptable_matches):
        mock_acceptable_matches.return_value = [lambda x: x.cert_path]
        test_cert_path = os.path.join(self.config.config_dir, 'live', 'example.org',
                'cert.pem')
        self.config.cert_path = test_cert_path
        assert 'example.org' == self._call(self.config)

    @mock.patch('certbot._internal.cert_manager._acceptable_matches')
    def test_options_archive_cert(self, mock_acceptable_matches):
        # Also this and the next test check that the regex of _archive_files is working.
        self.config.cert_path = os.path.join(self.config.config_dir, 'archive', 'example.org',
            'cert11.pem')
        mock_acceptable_matches.return_value = [lambda x: self._archive_files(x, 'cert')]
        assert 'example.org' == self._call(self.config)

    @mock.patch('certbot._internal.cert_manager._acceptable_matches')
    def test_options_archive_fullchain(self, mock_acceptable_matches):
        self.config.cert_path = os.path.join(self.config.config_dir, 'archive',
            'example.org', 'fullchain11.pem')
        mock_acceptable_matches.return_value = [lambda x:
                self._archive_files(x, 'fullchain')]
        assert 'example.org' == self._call(self.config)

    def test_only_path(self):
        self.config.cert_path = self.fullchain
        assert 'example.org' == self._call(self.config)


class MatchAndCheckOverlaps(storage_test.BaseRenewableCertTest):
    """Tests for certbot._internal.cert_manager.match_and_check_overlaps w/o overlapping
       archive dirs."""
    # A test with real overlapping archive dirs can be found in tests/boulder_integration.sh
    def setUp(self):
        super().setUp()
        self.config_file.write()
        self._write_out_ex_kinds()
        self.fullchain = os.path.join(self.config.config_dir, 'live', 'example.org',
                'fullchain.pem')
        self.config.cert_path = self.fullchain

    def _call(self, cli_config, acceptable_matches, match_func, rv_func):
        from certbot._internal.cert_manager import match_and_check_overlaps
        return match_and_check_overlaps(cli_config, acceptable_matches, match_func, rv_func)

    def test_basic_match(self):
        from certbot._internal.cert_manager import _acceptable_matches
        assert ['example.org'] == self._call(self.config, _acceptable_matches(),
            lambda x: self.config.cert_path, lambda x: x.lineagename)

    @mock.patch('certbot._internal.cert_manager._search_lineages')
    def test_no_matches(self, mock_search_lineages):
        mock_search_lineages.return_value = []
        with pytest.raises(errors.Error):
            self._call(self.config, None, None, None)

    @mock.patch('certbot._internal.cert_manager._search_lineages')
    def test_too_many_matches(self, mock_search_lineages):
        mock_search_lineages.return_value = ['spider', 'dance']
        with pytest.raises(errors.OverlappingMatchFound):
            self._call(self.config, None, None, None)


class GetCertnameTest(unittest.TestCase):
    """Tests for certbot._internal.cert_manager."""

    def setUp(self):
        get_utility_patch = test_util.patch_display_util()
        self.mock_get_utility = get_utility_patch.start()
        self.addCleanup(get_utility_patch.stop)
        self.config = mock.MagicMock()
        self.config.certname = None

    @mock.patch('certbot._internal.storage.renewal_conf_files')
    @mock.patch('certbot._internal.storage.lineagename_for_filename')
    def test_get_certnames(self, mock_name, mock_files):
        mock_files.return_value = ['example.com.conf']
        mock_name.return_value = 'example.com'
        from certbot._internal import cert_manager
        prompt = "Which certificate would you"
        self.mock_get_utility().menu.return_value = (display_util.OK, 0)
        assert cert_manager.get_certnames(
                self.config, "verb", allow_multiple=False) == ['example.com']
        assert prompt in self.mock_get_utility().menu.call_args[0][0]

    @mock.patch('certbot._internal.storage.renewal_conf_files')
    @mock.patch('certbot._internal.storage.lineagename_for_filename')
    def test_get_certnames_custom_prompt(self, mock_name, mock_files):
        mock_files.return_value = ['example.com.conf']
        mock_name.return_value = 'example.com'
        from certbot._internal import cert_manager
        prompt = "custom prompt"
        self.mock_get_utility().menu.return_value = (display_util.OK, 0)
        assert cert_manager.get_certnames(
                self.config, "verb", allow_multiple=False, custom_prompt=prompt) == \
            ['example.com']
        assert self.mock_get_utility().menu.call_args[0][0] == \
                          prompt

    @mock.patch('certbot._internal.storage.renewal_conf_files')
    @mock.patch('certbot._internal.storage.lineagename_for_filename')
    def test_get_certnames_user_abort(self, mock_name, mock_files):
        mock_files.return_value = ['example.com.conf']
        mock_name.return_value = 'example.com'
        from certbot._internal import cert_manager
        self.mock_get_utility().menu.return_value = (display_util.CANCEL, 0)
        with pytest.raises(errors.Error):
            cert_manager.get_certnames(self.config, "erroring_anyway", allow_multiple=False)

    @mock.patch('certbot._internal.storage.renewal_conf_files')
    @mock.patch('certbot._internal.storage.lineagename_for_filename')
    def test_get_certnames_allow_multiple(self, mock_name, mock_files):
        mock_files.return_value = ['example.com.conf']
        mock_name.return_value = 'example.com'
        from certbot._internal import cert_manager
        prompt = "Which certificate(s) would you"
        self.mock_get_utility().checklist.return_value = (display_util.OK,
                                                          ['example.com'])
        assert cert_manager.get_certnames(
                self.config, "verb", allow_multiple=True) == ['example.com']
        assert prompt in self.mock_get_utility().checklist.call_args[0][0]

    @mock.patch('certbot._internal.storage.renewal_conf_files')
    @mock.patch('certbot._internal.storage.lineagename_for_filename')
    def test_get_certnames_allow_multiple_custom_prompt(self, mock_name, mock_files):
        mock_files.return_value = ['example.com.conf']
        mock_name.return_value = 'example.com'
        from certbot._internal import cert_manager
        prompt = "custom prompt"
        self.mock_get_utility().checklist.return_value = (display_util.OK,
                                                          ['example.com'])
        assert cert_manager.get_certnames(
                self.config, "verb", allow_multiple=True, custom_prompt=prompt) == \
            ['example.com']
        assert self.mock_get_utility().checklist.call_args[0][0] == \
            prompt

    @mock.patch('certbot._internal.storage.renewal_conf_files')
    @mock.patch('certbot._internal.storage.lineagename_for_filename')
    def test_get_certnames_allow_multiple_user_abort(self, mock_name, mock_files):
        mock_files.return_value = ['example.com.conf']
        mock_name.return_value = 'example.com'
        from certbot._internal import cert_manager
        self.mock_get_utility().checklist.return_value = (display_util.CANCEL, [])
        with pytest.raises(errors.Error):
            cert_manager.get_certnames(self.config, "erroring_anyway", allow_multiple=True)


if __name__ == "__main__":
    sys.exit(pytest.main(sys.argv[1:] + [__file__]))  # pragma: no cover



"""Tests for new style enhancements"""
import sys
from unittest import mock

import pytest

from certbot._internal.plugins import null
from certbot.plugins import enhancements
import certbot.tests.util as test_util


class EnhancementTest(test_util.ConfigTestCase):
    """Tests for new style enhancements in certbot.plugins.enhancements"""

    def setUp(self):
        super().setUp()
        self.mockinstaller = mock.MagicMock(spec=enhancements.AutoHSTSEnhancement)

    @test_util.patch_display_util()
    def test_enhancement_enabled_enhancements(self, _):
        FAKEINDEX = [
            {
                "name": "autohsts",
                "cli_dest": "auto_hsts",
            },
            {
                "name": "somethingelse",
                "cli_dest": "something",
            }
        ]
        with mock.patch("certbot.plugins.enhancements._INDEX", FAKEINDEX):
            self.config.auto_hsts = True
            self.config.something = True
            enabled = list(enhancements.enabled_enhancements(self.config))
        assert len(enabled) == 2
        assert [i for i in enabled if i["name"] == "autohsts"]
        assert [i for i in enabled if i["name"] == "somethingelse"]

    def test_are_requested(self):
        assert len(list(enhancements.enabled_enhancements(self.config))) == 0
        assert not enhancements.are_requested(self.config)
        self.config.auto_hsts = True
        assert len(list(enhancements.enabled_enhancements(self.config))) == 1
        assert enhancements.are_requested(self.config)

    def test_are_supported(self):
        self.config.auto_hsts = True
        unsupported = null.Installer(self.config, "null")
        assert enhancements.are_supported(self.config, self.mockinstaller)
        assert not enhancements.are_supported(self.config, unsupported)

    def test_enable(self):
        self.config.auto_hsts = True
        domains = ["example.com", "www.example.com"]
        lineage = "lineage"
        enhancements.enable(lineage, domains, self.mockinstaller, self.config)
        assert self.mockinstaller.enable_autohsts.called
        assert self.mockinstaller.enable_autohsts.call_args[0] == (lineage, domains)


if __name__ == '__main__':
    sys.exit(pytest.main(sys.argv[1:] + [__file__]))  # pragma: no cover



"""Test for certbot._internal.plugins.nginx.nginxparser."""
import copy
import operator
import sys
import tempfile
import unittest

from pyparsing import ParseException
import pytest

from certbot._internal.plugins.nginx.nginxparser import dump
from certbot._internal.plugins.nginx.nginxparser import dumps
from certbot._internal.plugins.nginx.nginxparser import load
from certbot._internal.plugins.nginx.nginxparser import loads
from certbot._internal.plugins.nginx.nginxparser import RawNginxParser
from certbot._internal.plugins.nginx.nginxparser import UnspacedList
from certbot._internal.tests.plugins.nginx import test_util as util

FIRST = operator.itemgetter(0)


class TestRawNginxParser(unittest.TestCase):
    """Test the raw low-level Nginx config parser."""

    def test_assignments(self):
        parsed = RawNginxParser.assignment.parse_string('root /test;').asList()
        assert parsed == ['root', ' ', '/test']
        parsed = RawNginxParser.assignment.parse_string('root /test;foo bar;').asList()
        assert parsed == ['root', ' ', '/test'], ['foo', ' ', 'bar']

    def test_blocks(self):
        parsed = RawNginxParser.block.parse_string('foo {}').asList()
        assert parsed == [['foo', ' '], []]
        parsed = RawNginxParser.block.parse_string('location /foo{}').asList()
        assert parsed == [['location', ' ', '/foo'], []]
        parsed = RawNginxParser.block.parse_string('foo { bar foo ; }').asList()
        assert parsed == [['foo', ' '], [[' ', 'bar', ' ', 'foo', ' '], ' ']]

    def test_nested_blocks(self):
        parsed = RawNginxParser.block.parse_string('foo { bar {} }').asList()
        block, content = parsed
        assert FIRST(content) == [[' ', 'bar', ' '], []]
        assert FIRST(block) == 'foo'

    def test_dump_as_string(self):
        dumped = dumps(UnspacedList([
            ['user', ' ', 'www-data'],
            [['\n', 'server', ' '], [
                ['\n    ', 'listen', ' ', '80'],
                ['\n    ', 'server_name', ' ', 'foo.com'],
                ['\n    ', 'root', ' ', '/home/ubuntu/sites/foo/'],
                [['\n\n    ', 'location', ' ', '/status', ' '], [
                    ['\n        ', 'check_status', ''],
                    [['\n\n        ', 'types', ' '],
                    [['\n            ', 'image/jpeg', ' ', 'jpg']]],
                ]]
            ]]]))

        assert dumped.split('\n') == \
                         'user www-data;\n' \
                         'server {\n' \
                         '    listen 80;\n' \
                         '    server_name foo.com;\n' \
                         '    root /home/ubuntu/sites/foo/;\n' \
                         '\n' \
                         '    location /status {\n' \
                         '        check_status;\n' \
                         '\n' \
                         '        types {\n' \
                         '            image/jpeg jpg;}}}'.split('\n')

    def test_parse_from_file(self):
        with util.get_data_filename('foo.conf') as path:
            with open(path) as handle:
                parsed = util.filter_comments(load(handle))
        assert parsed == \
            [['user', 'www-data'],
            [['http'],
            [[['server'], [
                ['listen', '*:80', 'default_server', 'ssl'],
                ['server_name', '*.www.foo.com', '*.www.example.com'],
                ['root', '/home/ubuntu/sites/foo/'],
                [['location', '/status'], [
                    [['types'], [['image/jpeg', 'jpg']]],
                ]],
                [['location', '~', r'case_sensitive\.php$'], [
                    ['index', 'index.php'],
                    ['root', '/var/root'],
                ]],
                [['location', '~*', r'case_insensitive\.php$'], []],
                [['location', '=', r'exact_match\.php$'], []],
                [['location', '^~', r'ignore_regex\.php$'], []]
            ]]]]]

    def test_parse_from_file2(self):
        with util.get_data_filename('edge_cases.conf') as path:
            with open(path) as handle:
                parsed = util.filter_comments(load(handle))
        assert parsed == \
            [[['server'], [['server_name', 'simple']]],
            [['server'],
            [['server_name', 'with.if'],
            [['location', '~', '^/services/.+$'],
                [[['if', '($request_filename', '~*', '\\.(ttf|woff)$)'],
                [['add_header', 'Access-Control-Allow-Origin', '"*"']]]]]]],
            [['server'],
            [['server_name', 'with.complicated.headers'],
            [['location', '~*', '\\.(?:gif|jpe?g|png)$'],
                [['add_header', 'Pragma', 'public'],
                ['add_header',
                'Cache-Control', '\'public, must-revalidate, proxy-revalidate\'',
                '"test,;{}"', 'foo'],
                ['blah', '"hello;world"'],
                ['try_files', '$uri', '@rewrites']]]]]]

    def test_parse_from_file3(self):
        with util.get_data_filename('multiline_quotes.conf') as path:
            with open(path) as handle:
                parsed = util.filter_comments(load(handle))
        assert parsed == \
            [[['http'],
                [[['server'],
                    [['listen', '*:443'],
                    [['location', '/'],
                        [['body_filter_by_lua',
                          '\'ngx.ctx.buffered = (ngx.ctx.buffered or "")'
                          ' .. string.sub(ngx.arg[1], 1, 1000)\n'
                          '                            '
                          'if ngx.arg[2] then\n'
                          '                              '
                          'ngx.var.resp_body = ngx.ctx.buffered\n'
                          '                            end\'']]]]]]]]

    def test_abort_on_parse_failure(self):
        with util.get_data_filename('broken.conf') as path:
            with open(path) as handle:
                with pytest.raises(ParseException):
                    load(handle)

    def test_dump_as_file(self):
        with util.get_data_filename('nginx.conf') as path:
            with open(path) as handle:
                parsed = load(handle)
        parsed[-1][-1].append(UnspacedList([['server'],
                               [['listen', ' ', '443', ' ', 'ssl'],
                                ['server_name', ' ', 'localhost'],
                                ['ssl_certificate', ' ', 'cert.pem'],
                                ['ssl_certificate_key', ' ', 'cert.key'],
                                ['ssl_session_cache', ' ', 'shared:SSL:1m'],
                                ['ssl_session_timeout', ' ', '5m'],
                                ['ssl_ciphers', ' ', 'HIGH:!aNULL:!MD5'],
                                [['location', ' ', '/'],
                                 [['root', ' ', 'html'],
                                  ['index', ' ', 'index.html', ' ', 'index.htm']]]]]))

        with tempfile.TemporaryFile(mode='w+t') as f:
            dump(parsed, f)
            f.seek(0)
            parsed_new = load(f)
        assert parsed == parsed_new

    def test_comments(self):
        with util.get_data_filename('minimalistic_comments.conf') as path:
            with open(path) as handle:
                parsed = load(handle)

        with tempfile.TemporaryFile(mode='w+t') as f:
            dump(parsed, f)
            f.seek(0)
            parsed_new = load(f)

        assert parsed == parsed_new
        assert parsed_new == [
            ['#', " Use bar.conf when it's a full moon!"],
            ['include', 'foo.conf'],
            ['#', ' Kilroy was here'],
            ['check_status'],
            [['server'],
             [['#', ''],
              ['#', " Don't forget to open up your firewall!"],
              ['#', ''],
              ['listen', '1234'],
              ['#', ' listen 80;']]],
        ]

    def test_issue_518(self):
        parsed = loads('if ($http_accept ~* "webp") { set $webp "true"; }')

        assert parsed == [
            [['if', '($http_accept', '~*', '"webp")'],
             [['set', '$webp', '"true"']]]
        ]

    def test_comment_in_block(self):
        parsed = loads("""http {
          # server{
          }""")

        assert parsed == [
            [['http'],
             [['#', ' server{']]]
        ]

    def test_comment_inside_multiline_directive(self):
        # See https://github.com/certbot/certbot/issues/10598
        # An unbalanced double-quote inside a comment that sits between tokens
        # of a multi-line directive must not be sucked into a multiline quoted
        # string and run away across the rest of the file.
        source = ("http {\n"
                  "    log_format json_test escape=json '{'\n"
                  "        #  '\"req\": \"$request\"'  # GET \"/test?...\n"
                  "    '}';\n"
                  "}\n")
        parsed = loads(source)
        # The config parses and round-trips byte-for-byte.
        assert dumps(parsed) == source

    def test_access_log(self):
        # see issue #3798
        parsed = loads('access_log syslog:server=unix:/dev/log,facility=auth,'
            'tag=nginx_post,severity=info custom;')

        assert parsed == [
            ['access_log',
             'syslog:server=unix:/dev/log,facility=auth,tag=nginx_post,severity=info',
             'custom']
        ]

    def test_add_header(self):
        # see issue #3798
        parsed = loads('add_header Cache-Control no-cache,no-store,must-revalidate,max-age=0;')

        assert parsed == [
            ['add_header', 'Cache-Control', 'no-cache,no-store,must-revalidate,max-age=0']
        ]

    def test_map_then_assignment_in_block(self):
        # see issue #3798
        test_str = """http {
            map $http_upgrade $connection_upgrade {
              default upgrade;
              ''      close;
              "~Opera Mini" 1;
              *.example.com 1;
            }
            one;
        }"""
        parsed = loads(test_str)
        assert parsed == [
            [['http'], [
                [['map', '$http_upgrade', '$connection_upgrade'], [
                    ['default', 'upgrade'],
                    ["''", 'close'],
                    ['"~Opera Mini"', '1'],
                    ['*.example.com', '1']
                ]],
                ['one']
            ]]
        ]

    def test_variable_name(self):
        parsed = loads('try_files /typo3temp/tx_ncstaticfilecache/'
            '$host${request_uri}index.html @nocache;')

        assert parsed == [
            ['try_files',
             '/typo3temp/tx_ncstaticfilecache/$host${request_uri}index.html',
             '@nocache']
        ]

    def test_weird_blocks(self):
        test = r"""
            if ($http_user_agent ~ MSIE) {
                rewrite ^(.*)$ /msie/$1 break;
            }

            if ($http_cookie ~* "id=([^;]+)(?:;|$)") {
               set $id $1;
            }

            if ($request_method = POST) {
               return 405;
            }

            if ($request_method) {
               return 403;
            }

            if ($args ~ post=140){
              rewrite ^ http://example.com/;
            }

            location ~ ^/users/(.+\.(?:gif|jpe?g|png))$ {
              alias /data/w3/images/$1;
            }

            proxy_set_header X-Origin-URI ${scheme}://${http_host}/$request_uri;
        """
        parsed = loads(test)
        assert parsed == [[['if', '($http_user_agent', '~', 'MSIE)'],
            [['rewrite', '^(.*)$', '/msie/$1', 'break']]],
            [['if', '($http_cookie', '~*', '"id=([^;]+)(?:;|$)")'], [['set', '$id', '$1']]],
            [['if', '($request_method', '=', 'POST)'], [['return', '405']]],
            [['if', '($request_method)'],
            [['return', '403']]], [['if', '($args', '~', 'post=140)'],
            [['rewrite', '^', 'http://example.com/']]],
            [['location', '~', '^/users/(.+\\.(?:gif|jpe?g|png))$'],
            [['alias', '/data/w3/images/$1']]],
            ['proxy_set_header', 'X-Origin-URI', '${scheme}://${http_host}/$request_uri']]

    def test_edge_cases(self):
        # quotes
        parsed = loads(r'"hello\""; # blah "heh heh"')
        assert parsed == [['"hello\\""'], ['#', ' blah "heh heh"']]

        # if with comment
        parsed = loads("""if ($http_cookie ~* "id=([^;]+)(?:;|$)") { # blah )
            }""")
        assert parsed == [[['if', '($http_cookie', '~*', '"id=([^;]+)(?:;|$)")'],
            [['#', ' blah )']]]]

        # end paren
        test = """
            one"test";
            ("two");
            "test")red;
            "test")"blue";
            "test")"three;
            (one"test")one;
            one";
            one"test;
            one"test"one;
        """
        parsed = loads(test)
        assert parsed == [
            ['one"test"'],
            ['("two")'],
            ['"test")red'],
            ['"test")"blue"'],
            ['"test")"three'],
            ['(one"test")one'],
            ['one"'],
            ['one"test'],
            ['one"test"one']
        ]
        with pytest.raises(ParseException):
            loads(r'"test"one;') # fails
        with pytest.raises(ParseException):
            loads(r'"test;') # fails

        # newlines
        test = """
            server_name foo.example.com bar.example.com \
                        baz.example.com qux.example.com;
            server_name foo.example.com bar.example.com
                        baz.example.com qux.example.com;
        """
        parsed = loads(test)
        assert parsed == [
            ['server_name', 'foo.example.com', 'bar.example.com',
                'baz.example.com', 'qux.example.com'],
            ['server_name', 'foo.example.com', 'bar.example.com',
                'baz.example.com', 'qux.example.com']
        ]

        # variable weirdness
        parsed = loads("directive $var ${var} $ ${};")
        assert parsed == [['directive', '$var', '${var}', '$', '${}']]
        with pytest.raises(ParseException):
            loads("server {server_name test.com};")
        assert loads("blag${dfgdfg};") == [['blag${dfgdfg}']]
        with pytest.raises(ParseException):
            loads("blag${dfgdf{g};")

        # empty file
        parsed = loads("")
        assert parsed == []

    def test_non_breaking_spaces(self):
        # non-breaking spaces
        test = u'\u00a0'
        loads(test)
        test = """
        map $http_upgrade $connection_upgrade {
            default upgrade;
            ''      close;
        }
        """
        loads(test)

    def test_location_comment_issue(self):
        # See discussion at https://github.com/certbot/certbot/issues/10264
        already_good = '''
        location = /resume
        # x
        { rewrite .* /Files/Adam_Lein_resume.pdf redirect; }
        '''
        loads(already_good)
        already_good = '''
        location = /resume
        { rewrite .* /Files/Adam_Lein_resume.pdf redirect; }
        # {
        '''
        loads(already_good)
        needs_fixing = '''
        location = /resume
        # {
        { rewrite .* /Files/Adam_Lein_resume.pdf redirect; }
        '''
        with pytest.raises(ParseException):
            loads(needs_fixing) # fails
        needs_fixing = '''
        location = /resume
        # x{
        { rewrite .* /Files/Adam_Lein_resume.pdf redirect; }
        '''
        with pytest.raises(ParseException):
            loads(needs_fixing) # fails
        needs_fixing = '''
        location = /resume
        #{
        { rewrite .* /Files/Adam_Lein_resume.pdf redirect; }
        '''
        with pytest.raises(ParseException):
            loads(needs_fixing) # fails
        needs_fixing = '''
        location = /resume
        # {x
        { rewrite .* /Files/Adam_Lein_resume.pdf redirect; }
        '''
        with pytest.raises(ParseException):
            loads(needs_fixing) # fails


class TestUnspacedList(unittest.TestCase):
    """Test the UnspacedList data structure"""
    def setUp(self):
        self.a = ["\n    ", "things", " ", "quirk"]
        self.b = ["y", " "]
        self.l = self.a[:]
        self.l2 = self.b[:]
        self.ul = UnspacedList(self.l)
        self.ul2 = UnspacedList(self.l2)

    def test_construction(self):
        assert self.ul == ["things", "quirk"]
        assert self.ul2 == ["y"]

    def test_append(self):
        ul3 = copy.deepcopy(self.ul)
        ul3.append("wise")
        assert ul3 == ["things", "quirk", "wise"]
        assert ul3.spaced == self.a + ["wise"]

    def test_add(self):
        ul3 = self.ul + self.ul2
        assert ul3 == ["things", "quirk", "y"]
        assert ul3.spaced == self.a + self.b
        assert self.ul.spaced == self.a
        ul3 = self.ul + self.l2
        assert ul3 == ["things", "quirk", "y"]
        assert ul3.spaced == self.a + self.b

    def test_extend(self):
        ul3 = copy.deepcopy(self.ul)
        ul3.extend(self.ul2)
        assert ul3 == ["things", "quirk", "y"]
        assert ul3.spaced == self.a + self.b
        assert self.ul.spaced == self.a

    def test_set(self):
        ul3 = copy.deepcopy(self.ul)
        ul3[0] = "zither"
        l = ["\n ", "zather", "zest"]
        ul3[1] = UnspacedList(l)
        assert ul3 == ["zither", ["zather", "zest"]]
        assert ul3.spaced == [self.a[0], "zither", " ", l]

    def test_get(self):
        with pytest.raises(IndexError):
            self.ul2.__getitem__(2)
        with pytest.raises(IndexError):
            self.ul2.__getitem__(-3)

    def test_insert(self):
        x = UnspacedList(
                [['\n    ', 'listen', '       ', '69.50.225.155:9000'],
                ['\n    ', 'listen', '       ', '127.0.0.1'],
                ['\n    ', 'server_name', ' ', '.example.com'],
                ['\n    ', 'server_name', ' ', 'example.*'], '\n',
                ['listen', ' ', '5001', ' ', 'ssl']])
        x.insert(5, "FROGZ")
        assert x == \
            [['listen', '69.50.225.155:9000'], ['listen', '127.0.0.1'],
            ['server_name', '.example.com'], ['server_name', 'example.*'],
            ['listen', '5001', 'ssl'], 'FROGZ']
        assert x.spaced == \
            [['\n    ', 'listen', '       ', '69.50.225.155:9000'],
            ['\n    ', 'listen', '       ', '127.0.0.1'],
            ['\n    ', 'server_name', ' ', '.example.com'],
            ['\n    ', 'server_name', ' ', 'example.*'], '\n',
            ['listen', ' ', '5001', ' ', 'ssl'],
            'FROGZ']

    def test_rawlists(self):
        ul3 = copy.deepcopy(self.ul)
        ul3.insert(0, "some")
        ul3.append("why")
        ul3.extend(["did", "whether"])
        del ul3[2]
        assert ul3 == ["some", "things", "why", "did", "whether"]

    def test_is_dirty(self):
        assert self.ul2.is_dirty() is False
        ul3 = UnspacedList([])
        ul3.append(self.ul)
        assert self.ul.is_dirty() is False
        assert ul3.is_dirty() is True
        ul4 = UnspacedList([[1], [2, 3, 4]])
        assert ul4.is_dirty() is False
        ul4[1][2] = 5
        assert ul4.is_dirty() is True


if __name__ == '__main__':
    sys.exit(pytest.main(sys.argv[1:] + [__file__]))  # pragma: no cover
